/*	$NetBSD: wd.c,v 1.335.2.1.4.1 2008/03/30 14:44:42 jdc Exp $ */

/*
 * Copyright (c) 1998, 2001 Manuel Bouyer.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *	notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *	notice, this list of conditions and the following disclaimer in the
 *	documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *	must display the following acknowledgement:
 *  This product includes software developed by Manuel Bouyer.
 * 4. The name of the author may not be used to endorse or promote products
 *	derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*-
 * Copyright (c) 1998, 2003, 2004 The NetBSD Foundation, Inc.
 * All rights reserved.
 *
 * This code is derived from software contributed to The NetBSD Foundation
 * by Charles M. Hannum and by Onno van der Linden.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *        This product includes software developed by the NetBSD
 *        Foundation, Inc. and its contributors.
 * 4. Neither the name of The NetBSD Foundation nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE NETBSD FOUNDATION, INC. AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FOUNDATION OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * 2009: modified by minoru murashima.
 */

#include <sys/types.h>
#include <sys/param.h>
#include <sys/limits.h>
#include <sys/systm.h>
#include <sys/syslog.h>
#include <sys/stat.h>
#include <sys/lock.h>
#include <sys/conf.h>
#include <sys/fcntl.h>
#include <sys/errno.h>
#include <sys/uio.h>
#include <sys/queue.h>
#include <sys/callout.h>
#include <sys/timevar.h>
#include <sys/device.h>
#include <sys/buf.h>
#include <sys/dkio.h>
#include <sys/disk.h>
#include <sys/ataio.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <GENERIC/locators.h>
#include <ata/atareg.h>
#include <ata/atavar.h>
#include <ata/wdvar.h>

//=====================================  ===================================================

#define	WD_QUIRK_SPLIT_MOD15_WRITE	0x0001		/* must split certain writes */
#define	WD_QUIRK_FORCE_LBA48		0x0002		/* must use LBA48 commands */
#define	LBA48_THRESHOLD				(0xfffffff)	/* 128GB / DEV_BSIZE */
#define	WDIORETRIES_SINGLE			4			/* number of retries before single-sector */
#define	WDIORETRIES					5			/* number of retries before giving up */
#define	RECOVERYTIME				hz/2		/* time to wait before retrying a cmd */
#define	WD_QUIRK_FMT				"\20\1SPLIT_MOD15_WRITE\2FORCE_LBA48"

#define	WDUNIT(dev)					DISKUNIT(dev)
#define	WDPART(dev)					DISKPART(dev)
#define	MAKEWDDEV(maj, unit, part)	MAKEDISKDEV(maj, unit, part)
#define	WDLABELDEV(dev)				(MAKEWDDEV(major(dev), WDUNIT(dev), RAW_PART))

struct wd_ioctl {
	LIST_ENTRY(wd_ioctl) wi_list;
	struct buf wi_bp;
	struct uio wi_uio;
	struct iovec wi_iov;
	atareq_t wi_atareq;
	struct wd_softc *wi_softc;
};

//===================================== Х륤ݡ =======================================

extern struct cfdriver wd_cd;
extern const struct cdevsw wd_cdevsw;

//===================================== PRIVATE ====================================================

static const char at_errbits[] = "\20\10ERROR\11TIMEOU\12DF";

LIST_HEAD(, wd_ioctl) wi_head;

/*
 * Quirk table for IDE drives.  Put more-specific matches first, since
 * a simple globbing routine is used for matching.
 */
static const struct wd_quirk {
	const char *wdq_match;		/* inquiry pattern to match */
	int wdq_quirks;				/* drive quirks */
} wd_quirk_table[] = {
	/*
	 * Some Seagate S-ATA drives have a PHY which can get confused
	 * with the way data is packetized by some S-ATA controllers.
	 *
	 * The work-around is to split in two any write transfer whose
	 * sector count % 15 == 1 (assuming 512 byte sectors).
	 *
	 * XXX This is an incomplete list.  There are at least a couple
	 * XXX more model numbers.  If you have trouble with such transfers
	 * XXX (8K is the most common) on Seagate S-ATA drives, please
	 * XXX notify thorpej@NetBSD.org.
	 */
	{ "ST3120023AS",
	  WD_QUIRK_SPLIT_MOD15_WRITE },
	{ "ST380023AS",
	  WD_QUIRK_SPLIT_MOD15_WRITE },

	/*
	 * These seagate drives seems to have issue addressing sector 0xfffffff
	 * (aka LBA48_THRESHOLD) in LBA mode. The workaround is to force
	 * LBA48
	 * Note that we can't just change the code to always use LBA48 for
	 * sector 0xfffffff, because this would break valid and working
	 * setups using LBA48 drives on non-LBA48-capable controllers
	 * (and it's hard to get a list of such controllers)
	 */
	{ "ST3160021A*",
	  WD_QUIRK_FORCE_LBA48 },
	{ "ST3160811A*",
	  WD_QUIRK_FORCE_LBA48 },
	{ "ST3160812A*",
	  WD_QUIRK_FORCE_LBA48 },
	{ "ST3160023A*",
	  WD_QUIRK_FORCE_LBA48 },
	{ "ST3160827A*",
	  WD_QUIRK_FORCE_LBA48 },
	/* Attempt to catch all seagate drives larger than 200GB */
	{ "ST3[2-9][0-9][0-9][0-9][0-9][0-9][A-Z]*",
	  WD_QUIRK_FORCE_LBA48 },
	{ NULL,
	  0 }
};

static void wdperror(
	const struct wd_softc *wd)
{
	static const char *const errstr0_3[] = {
		"address mark not found",
		"track 0 not found", 
		"aborted command", 
		"media change requested",
		"id not found", 
		"media changed", 
		"uncorrectable data error",
		"bad block detected"};
	static const char *const errstr4_5[] = {
		"obsolete (address mark not found)",
		"no media/write protected", 
		"aborted command",
		"media change requested", 
		"id not found", 
		"media changed",
		"uncorrectable data error", 
		"interface CRC error"};
	const char *const *errstr;
	int i;
	const char *sep = "";

	const char *devname = wd->sc_dev.dv_xname;
	struct ata_drive_datas *drvp = wd->drvp;
	int errno = wd->sc_wdc_bio.r_error;

	if (drvp->ata_vers >= 4) {
		errstr = errstr4_5;
	}
	else {
		errstr = errstr0_3;
	}

	printf("%s: (", devname);

	if (errno == 0) {
		printf("error not notified");
	}

	for (i = 0; i < 8; i++) {
		if (errno & (1 << i)) {
			printf("%s%s", sep, errstr[i]);
			sep = ", ";
		}
	}
	printf(")\n");
}

static void wdrestart(void *);

static void __wdstart(
	struct wd_softc *wd, 
	struct buf *bp)
{
	/*
	 * Deal with the "split mod15 write" quirk.  We just divide the
	 * transfer in two, doing the first half and then then second half
	 * with the same command opening.
	 *
	 * Note we MUST do this here, because we can't let insertion
	 * into the bufq cause the transfers to be re-merged.
	 */
//	if (__predict_false((wd->sc_quirks & WD_QUIRK_SPLIT_MOD15_WRITE) != 0 &&
//				(bp->b_flags & B_READ) == 0 &&
//				bp->b_bcount > 512 &&
//				((bp->b_bcount / 512) % 15) == 1)) {
//	if ((wd->sc_quirks & WD_QUIRK_SPLIT_MOD15_WRITE) != 0 &&
//		(bp->b_flags & B_READ) == 0 &&
//		bp->b_bcount > 512 &&
//		((bp->b_bcount / 512) % 15) == 1) {
//		struct buf *nbp;

		/* already at splbio */
//		nbp = getiobuf_nowait();
//		if (__predict_false(nbp == NULL)) {
//		if (nbp == NULL) {
			/* No memory -- fail the iop. */
//			bp->b_error = ENOMEM;
//			bp->b_flags |= B_ERROR;
//			bp->b_resid = bp->b_bcount;
//			biodone(bp);
//			wd->openings++;
//			return;
//		}
//		nbp->b_error = 0;
//		nbp->b_proc = bp->b_proc;
//		nbp->b_vp = NULLVP;
//		nbp->b_dev = bp->b_dev;
//		nbp->b_bcount = bp->b_bcount / 2;
//		nbp->b_bufsize = bp->b_bcount / 2;
//		nbp->b_data = bp->b_data;
//		nbp->b_blkno = bp->b_blkno;
//		nbp->b_rawblkno = bp->b_rawblkno;
//		nbp->b_flags = bp->b_flags | B_CALL;
//		nbp->b_iodone = wd_split_mod15_write;

		/* Put ptr to orig buf in b_private and use new buf */
//		nbp->b_private = bp;

//		BIO_COPYPRIO(nbp, bp);

//		bp = nbp;
//	}

	wd->sc_wdc_bio.blkno = bp->b_rawblkno;
	wd->sc_wdc_bio.blkdone = 0;
	wd->sc_bp = bp;
	wd->sc_wdc_bio.uio = bp->uio;
//****************************************************************************************
	wd->sc_wdc_bio.task = bp->task;
//****************************************************************************************

	/*
	 * If we're retrying, retry in single-sector mode. This will give us
	 * the sector number of the problem, and will eventually allow the
	 * transfer to succeed.
	 */
	if (WDIORETRIES_SINGLE <= wd->retries) {
		wd->sc_wdc_bio.flags = ATA_SINGLE;
	}
	else {
		wd->sc_wdc_bio.flags = 0;
	}
	if (wd->sc_flags & WDF_LBA48 && (LBA48_THRESHOLD < wd->sc_wdc_bio.blkno || (wd->sc_quirks & WD_QUIRK_FORCE_LBA48) != 0)) {
		wd->sc_wdc_bio.flags |= ATA_LBA48;
	}
	if (wd->sc_flags & WDF_LBA) {
		wd->sc_wdc_bio.flags |= ATA_LBA;
	}
	if (bp->b_flags & B_READ) {
		wd->sc_wdc_bio.flags |= ATA_READ;
	}
	wd->sc_wdc_bio.bcount = bp->b_bcount;
//	wd->sc_wdc_bio.databuf = bp->b_data;

	/* Instrumentation. */
	disk_busy(&wd->sc_dk);
	switch (wd->atabus->ata_bio(wd->drvp, &wd->sc_wdc_bio)) {
	case ATACMD_TRY_AGAIN:
		callout_reset(&wd->sc_restart_ch, hz, wdrestart, wd);
		panic("callout_reset not support");
		break;
	case ATACMD_QUEUED:
	case ATACMD_COMPLETE:
		break;
	default:
		panic("__wdstart: bad return code from ata_bio()");
	}
}

/*
 * Queue a drive for I/O.
 */
/*}
static void wdstart(
	void *arg,
	struct buf *bp)
{
	struct wd_softc *wd = arg;
	struct buf *bp = NULL;

	while (0 < wd->openings) {
*/		/* Is there a buf for us ? */
/*		if ((bp = BUFQ_GET(wd->sc_q)) == NULL) {
			return;
		}
*/
		/*
		 * Make the command. First lock the device
		 */
/*		wd->openings--;

		wd->retries = 0;
		__wdstart(wd, bp);
	}
}
*/
static void wdrestart(
	void *v)
{
	struct wd_softc *wd = v;
	struct buf *bp = wd->sc_bp;
	int s;

	s = splbio();
	__wdstart(v, bp);
	splx(s);
}

static void wddone(
	void *v)
{
	struct wd_softc *wd = v;
	struct buf *bp = wd->sc_bp;
	const char *errmsg;
	int do_perror = 0;
	int nblks;

	if (bp == NULL) {
		return;
	}
	bp->b_resid = wd->sc_wdc_bio.bcount;
	switch (wd->sc_wdc_bio.error) {
	case ERR_DMA:
		errmsg = "DMA error";
		goto retry;
	case ERR_DF:
		errmsg = "device fault";
		goto retry;
	case TIMEOUT:
		errmsg = "device timeout";
		goto retry;
	case ERR_RESET:
		errmsg = "channel reset";
		goto retry2;
	case ERROR:
		/* Don't care about media change bits */
		if (wd->sc_wdc_bio.r_error != 0 && (wd->sc_wdc_bio.r_error & ~(WDCE_MC | WDCE_MCR)) == 0) {
			goto noerror;
		}
		errmsg = "error";
		do_perror = 1;
		if ((wd->sc_wdc_bio.r_error & (WDCE_IDNF | WDCE_ABRT)) && (wd->sc_quirks & WD_QUIRK_FORCE_LBA48) == 0) {
			nblks = wd->sc_wdc_bio.bcount / wd->sc_dk.dk_label->d_secsize;
			/*
			 * If we get a "id not found" when crossing the
			 * LBA48_THRESHOLD, and the drive is larger than
			 * 128GB, then we can assume the drive has the
			 * LBA48 bug and we switch to LBA48.
			 */
			if (wd->sc_wdc_bio.blkno <= LBA48_THRESHOLD &&
				wd->sc_wdc_bio.blkno + nblks > LBA48_THRESHOLD &&
				wd->sc_capacity > LBA48_THRESHOLD + 1) {
				errmsg = "LBA48 bug";
				wd->sc_quirks |= WD_QUIRK_FORCE_LBA48;
				do_perror = 0;
				goto retry2;
			}
		}
retry:		/* Just reset and retry. Can we do more ? */
		(*wd->atabus->ata_reset_drive)(wd->drvp, AT_RST_NOCMD);
retry2:
		diskerr(bp, "wd", errmsg, LOG_PRINTF, wd->sc_wdc_bio.blkdone, wd->sc_dk.dk_label);
		if (wd->retries < WDIORETRIES) {
			printf(", retrying");
		}
		printf("\n");
		if (do_perror) {
			wdperror(wd);
		}
		if (wd->retries < WDIORETRIES) {
			wd->retries++;
			callout_reset(&wd->sc_restart_ch, RECOVERYTIME, wdrestart, wd);
			panic("callout_reset not support");
			return;
		}

		bp->b_flags |= B_ERROR;
		bp->b_error = EIO;
		break;
	case NOERROR:
noerror:
		if ((wd->sc_wdc_bio.flags & ATA_CORR) || 0 < wd->retries) {
			printf("%s: soft error (corrected)\n", wd->sc_dev.dv_xname);
		}
		break;
	case ERR_NODEV:
		bp->b_flags |= B_ERROR;
		bp->b_error = EIO;
		break;
	}

	disk_unbusy(&wd->sc_dk, (bp->b_bcount - bp->b_resid), (bp->b_flags & B_READ));
#if NRND > 0
	rnd_add_uint32(&wd->rnd_source, bp->b_blkno);
#endif
	/* XXX Yuck, but we don't want to increment openings in this case */
//	if (__predict_false((bp->b_flags & B_CALL) != 0 && bp->b_iodone == wd_split_mod15_write)) {
//		biodone(bp);
//	}
//	else {
//		biodone(bp);
		wd->openings++;
//	}
//	wdstart(wd);
//	__wdstart(wd, bp);
}

/*
 * Allocate space for a ioctl queue structure.  Mostly taken from
 * scsipi_ioctl.c
 */
struct wd_ioctl *wi_get(void)
{
	struct wd_ioctl *wi;
	int s;

	wi = malloc(sizeof(struct wd_ioctl));
	simple_lock_init(&wi->wi_bp.b_interlock);
	s = splbio();
	LIST_INSERT_HEAD(&wi_head, wi, wi_list);
	splx(s);
	return (wi);
}

/*
 * Free an ioctl structure and remove it from our list
 */
void wi_free(
	struct wd_ioctl *wi)
{
	int s;

	s = splbio();
	LIST_REMOVE(wi, wi_list);
	splx(s);
	free(wi);
}

static int wd_flushcache(
	struct wd_softc *wd, 
	int flags)
{
	struct ata_command *ata_c;
	int error = 0;

	/*
	 * WDCC_FLUSHCACHE is here since ATA-4, but some drives report
	 * only ATA-2 and still support it.
	 */
	if (wd->drvp->ata_vers < 4 && ((wd->sc_params.atap_cmd_set2 & WDC_CMD2_FC) == 0 || wd->sc_params.atap_cmd_set2 == 0xffff)) {
		return ENODEV;
	}
	
	ata_c = malloc(sizeof(*ata_c));
	if (ata_c == NULL) {
		return ENOMEM;
	}
	memset(ata_c, 0, sizeof(*ata_c));
	if ((wd->sc_params.atap_cmd2_en & ATA_CMD2_LBA48) != 0 && (wd->sc_params.atap_cmd2_en & ATA_CMD2_FCE) != 0) {
		ata_c->r_command = WDCC_FLUSHCACHE_EXT;
	}
	else {
		ata_c->r_command = WDCC_FLUSHCACHE;
	}
	ata_c->r_st_bmask = WDCS_DRDY;
	ata_c->r_st_pmask = WDCS_DRDY;
	ata_c->flags = flags;
	ata_c->timeout = 30000; /* 30s timeout */
	if (wd->atabus->ata_exec_command(wd->drvp, ata_c) != ATACMD_COMPLETE) {
		printf("%s: flush cache command didn't complete\n", wd->sc_dev.dv_xname);
		error = EIO;
		goto err;
	}
	if (ata_c->flags & AT_ERROR) {
		if (ata_c->r_error == WDCE_ABRT) { /* command not supported */
			error = ENODEV;
			goto err;
		}
	}
	if (ata_c->flags & (AT_ERROR | AT_TIMEOU | AT_DF)) {
		char sbuf[sizeof(at_errbits) + 64];
		bitmask_snprintf(ata_c->flags, at_errbits, sbuf, sizeof(sbuf));
		printf("%s: wd_flushcache: status=%s\n", wd->sc_dev.dv_xname, sbuf);
		error = EIO;
		goto err;
	}

err:
	free(ata_c);
	return error;
}

static int wd_get_params(
	struct wd_softc *wd, 
	u_int8_t flags, 
	struct ataparams *params)
{
	switch (wd->atabus->ata_get_params(wd->drvp, flags, params)) {
	case CMD_AGAIN:
		return 1;
	case CMD_ERR:
		/*
		 * We `know' there's a drive here; just assume it's old.
		 * This geometry is only used to read the MBR and print a
		 * (false) attach message.
		 */
		strncpy(params->atap_model, "ST506", sizeof params->atap_model);
		params->atap_config = ATA_CFG_FIXED;
		params->atap_cylinders = 1024;
		params->atap_heads = 8;
		params->atap_sectors = 17;
		params->atap_multi = 1;
		params->atap_capabilities1 = params->atap_capabilities2 = 0;
		wd->drvp->ata_vers = -1; /* Mark it as pre-ATA */
		/* FALLTHROUGH */
	case CMD_OK:
//		wd_params_to_properties(wd, params);
		return 0;
	default:
		panic("wd_get_params: bad return code from ata_get_params");
		/* NOTREACHED */
	}
}

static int wd_getcache(
	struct wd_softc *wd, 
	int *bitsp)
{
	struct ataparams params;

	if (wd_get_params(wd, AT_WAIT, &params) != 0) {
		return EIO;
	}
	if (params.atap_cmd_set1 == 0x0000 ||
		params.atap_cmd_set1 == 0xffff ||
		(params.atap_cmd_set1 & WDC_CMD1_CACHE) == 0) {
		*bitsp = 0;
		return 0;
	}
	*bitsp = DKCACHE_WCHANGE | DKCACHE_READ;
	if (params.atap_cmd1_en & WDC_CMD1_CACHE) {
		*bitsp |= DKCACHE_WRITE;
	}

	return 0;
}

static int wd_setcache(
	struct wd_softc *wd, 
	int bits)
{
	struct ataparams params;
	struct ata_command *ata_c;
	int error = 0;

	if (wd_get_params(wd, AT_WAIT, &params) != 0) {
		return EIO;
	}

	if (params.atap_cmd_set1 == 0x0000 ||
		params.atap_cmd_set1 == 0xffff ||
		(params.atap_cmd_set1 & WDC_CMD1_CACHE) == 0) {
		return EOPNOTSUPP;
	}

	if ((bits & DKCACHE_READ) == 0 || (bits & DKCACHE_SAVE) != 0) {
		return EOPNOTSUPP;
	}

	ata_c = malloc(sizeof(*ata_c));
	if (ata_c == NULL) {
		return ENOMEM;
	}
	memset(ata_c, 0, sizeof(*ata_c));
	ata_c->r_command = SET_FEATURES;
	ata_c->r_st_bmask = 0;
	ata_c->r_st_pmask = 0;
	ata_c->timeout = 30000; /* 30s timeout */
	ata_c->flags = AT_WAIT;
	if (bits & DKCACHE_WRITE) {
		ata_c->r_features = WDSF_WRITE_CACHE_EN;
	}
	else {
		ata_c->r_features = WDSF_WRITE_CACHE_DS;
	}
	if (wd->atabus->ata_exec_command(wd->drvp, ata_c) != ATACMD_COMPLETE) {
		printf("%s: wd_setcache command not complete\n", wd->sc_dev.dv_xname);
		error = EIO;
		goto err;
	}
	if (ata_c->flags & (AT_ERROR | AT_TIMEOU | AT_DF)) {
		char sbuf[sizeof(at_errbits) + 64];
		bitmask_snprintf(ata_c->flags, at_errbits, sbuf, sizeof(sbuf));
		printf("%s: wd_setcache: status=%s\n", wd->sc_dev.dv_xname, sbuf);
		error = EIO;
		goto err;
	}

err:
	free(ata_c);
	return error;
}

static void wdgetdefaultlabel(
	struct wd_softc *wd, 
	struct disklabel *lp)
{
	memset(lp, 0, sizeof(struct disklabel));

	lp->d_secsize = DEV_BSIZE;
	lp->d_ntracks = wd->sc_params.atap_heads;
	lp->d_nsectors = wd->sc_params.atap_sectors;
	lp->d_ncylinders = 
		(wd->sc_flags & WDF_LBA) ? wd->sc_capacity / (wd->sc_params.atap_heads * wd->sc_params.atap_sectors) : wd->sc_params.atap_cylinders;
	lp->d_secpercyl = lp->d_ntracks * lp->d_nsectors;

	if (strcmp(wd->sc_params.atap_model, "ST506") == 0) {
		lp->d_type = DTYPE_ST506;
	}
	else {
		lp->d_type = DTYPE_ESDI;
	}

	strncpy(lp->d_typename, wd->sc_params.atap_model, 16);
	strncpy(lp->d_packname, "fictitious", 16);
	if (UINT32_MAX < wd->sc_capacity) {
		lp->d_secperunit = UINT32_MAX;
	}
	else {
		lp->d_secperunit = wd->sc_capacity;
	}
	lp->d_rpm = 3600;
	lp->d_interleave = 1;
	lp->d_flags = 0;

	lp->d_partitions[RAW_PART].p_offset = 0;
	lp->d_partitions[RAW_PART].p_size = lp->d_secperunit * (lp->d_secsize / DEV_BSIZE);
	lp->d_partitions[RAW_PART].p_fstype = FS_UNUSED;
	lp->d_npartitions = RAW_PART + 1;

	lp->d_magic = DISKMAGIC;
	lp->d_magic2 = DISKMAGIC;
	lp->d_checksum = dkcksum(lp);
}

/*
 * Fabricate a default disk label, and try to read the correct one.
 */
static void wdgetdisklabel(
	struct wd_softc *wd)
{
	struct disklabel *lp = wd->sc_dk.dk_label;
//	const char *errstring;
	int s;

//	memset(wd->sc_dk.dk_cpulabel, 0, sizeof(struct cpu_disklabel));

	wdgetdefaultlabel(wd, lp);

//	wd->sc_badsect[0] = -1;

	if (wd->drvp->state > RESET) {
		s = splbio();
		wd->drvp->drive_flags |= DRIVE_RESET;
		splx(s);
	}
//	errstring = readdisklabel(MAKEWDDEV(0, device_unit(&wd->sc_dev), RAW_PART), wdstrategy, lp, wd->sc_dk.dk_cpulabel);
//	if (errstring) {
		/*
		 * This probably happened because the drive's default
		 * geometry doesn't match the DOS geometry.  We
		 * assume the DOS geometry is now in the label and try
		 * again.  XXX This is a kluge.
		 */
//		if (wd->drvp->state > RESET) {
//			s = splbio();
//			wd->drvp->drive_flags |= DRIVE_RESET;
//			splx(s);
//		}
//		errstring = readdisklabel(MAKEWDDEV(0, device_unit(&wd->sc_dev), RAW_PART), wdstrategy, lp, wd->sc_dk.dk_cpulabel);
//	}
//	if (errstring) {
//		printf("%s: %s\n", wd->sc_dev.dv_xname, errstring);
//		return;
//	}

//	if (wd->drvp->state > RESET) {
//		s = splbio();
//		wd->drvp->drive_flags |= DRIVE_RESET;
//		splx(s);
//	}
}

static const struct wd_quirk *wd_lookup_quirks(
	const char *name)
{
	const struct wd_quirk *wdq;
	const char *estr;

	for (wdq = wd_quirk_table; wdq->wdq_match != NULL; wdq++) {
		/*
		 * We only want exact matches (which include matches
		 * against globbing characters).
		 */
		if (pmatch(name, wdq->wdq_match, &estr) == 2) {
			return (wdq);
		}
	}
	return (NULL);
}

/*
 * Find a wd_ioctl structure based on the struct buf.
 */
/*static struct wd_ioctl *wi_find(
	struct buf *bp)
{
	struct wd_ioctl *wi;
	int s;

	s = splbio();
	for (wi = wi_head.lh_first; wi != 0; wi = wi->wi_list.le_next) {
		if (bp == &wi->wi_bp) {
			break;
		}
	}
	splx(s);
	return (wi);
}
*/
/*
 * Read/write routine for a buffer.  Validates the arguments and schedules the
 * transfer.  Does not wait for the transfer to complete.
 */
static void wdstrategy(
	struct buf *bp)
{
	struct wd_softc *wd = device_lookup(&wd_cd, WDUNIT(bp->b_dev));
	struct disklabel *lp = wd->sc_dk.dk_label;
	daddr_t blkno;
	int s;

	/* Valid request?  */
	if (bp->b_blkno < 0 ||
		(bp->b_bcount % lp->d_secsize) != 0 ||
		(1 << NBBY) <= (bp->b_bcount / lp->d_secsize)) {
		bp->b_error = EINVAL;
		goto bad;
	}

	/* If device invalidated (e.g. media change, door open), error. */
	if ((wd->sc_flags & WDF_LOADED) == 0) {
		bp->b_error = EIO;
		goto bad;
	}

	/* If it's a null transfer, return immediately. */
	if (bp->b_bcount == 0) {
		goto done;
	}

	/*
	 * Do bounds checking, adjust transfer. if error, process.
	 * If end of partition, just return.
	 */
	if (WDPART(bp->b_dev) == RAW_PART) {
		if (bounds_check_with_mediasize(bp, DEV_BSIZE, wd->sc_capacity) <= 0) {
			goto done;
		}
	} 
	else {
		if (bounds_check_with_label(&wd->sc_dk, bp, (wd->sc_flags & (WDF_WLABEL|WDF_LABELLING)) != 0) <= 0) {
			goto done;
		}
	}

	/*
	 * Now convert the block number to absolute and put it in
	 * terms of the device's logical block size.
	 */
	if (DEV_BSIZE <= lp->d_secsize) {
		blkno = bp->b_blkno / (lp->d_secsize / DEV_BSIZE);
	}
	else {
		blkno = bp->b_blkno * (DEV_BSIZE / lp->d_secsize);
	}
	if (WDPART(bp->b_dev) != RAW_PART) {
		blkno += lp->d_partitions[WDPART(bp->b_dev)].p_offset;
	}
	bp->b_rawblkno = blkno;

	/* Queue transfer on drive, activate drive and controller if idle. */
	s = splbio();
//	BUFQ_PUT(wd->sc_q, bp);
//	wdstart(wd, bp);
//**************************************************************************************************
	wd->retries = 0;
	__wdstart(wd, bp);
//**************************************************************************************************
	splx(s);
	return;
bad:
	bp->b_flags |= B_ERROR;
done:
	/* Toss transfer; we're done early. */
	bp->b_resid = bp->b_bcount;
//	biodone(bp);
}

static int wdopen(
	dev_t dev, 
	int flag, 
	int fmt, 
	struct lwp *l)
{
	struct wd_softc *wd;
	int part, error;

	wd = device_lookup(&wd_cd, WDUNIT(dev));
	if (wd == NULL) {
		return (-ENXIO);
	}

	if (! device_is_active(&wd->sc_dev)) {
		return (-ENODEV);
	}

	part = WDPART(dev);

	if ((error = lockmgr(&wd->sc_dk.dk_openlock, LK_EXCLUSIVE, NULL)) != 0) {
		return (-error);
	}

	/*
	 * If there are wedges, and this is not RAW_PART, then we
	 * need to fail.
	 */
	if (wd->sc_dk.dk_nwedges != 0 && part != RAW_PART) {
		error = EBUSY;
		goto bad1;
	}

	/*
	 * If this is the first open of this device, add a reference
	 * to the adapter.
	 */
	if (wd->sc_dk.dk_openmask == 0 && (error = wd->atabus->ata_addref(wd->drvp)) != 0) {
		goto bad1;
	}

	if (wd->sc_dk.dk_openmask != 0) {
		/*
		 * If any partition is open, but the disk has been invalidated,
		 * disallow further opens.
		 */
		if ((wd->sc_flags & WDF_LOADED) == 0) {
			error = EIO;
			goto bad2;
		}
	} 
	else {
		if ((wd->sc_flags & WDF_LOADED) == 0) {
			wd->sc_flags |= WDF_LOADED;

			/* Load the physical device parameters. */
			wd_get_params(wd, AT_WAIT, &wd->sc_params);

			/* Load the partition info if not already loaded. */
			wdgetdisklabel(wd);
		}
	}

	/* Check that the partition exists. */
//	if (part != RAW_PART &&
//		(part >= wd->sc_dk.dk_label->d_npartitions ||
//		 wd->sc_dk.dk_label->d_partitions[part].p_fstype == FS_UNUSED)) {
//		error = ENXIO;
//		goto bad2;
//	}

	/* Insure only one open at a time. */
	switch (fmt) {
	case S_IFCHR:
		wd->sc_dk.dk_copenmask |= (1 << part);
		break;
	case S_IFBLK:
		wd->sc_dk.dk_bopenmask |= (1 << part);
		break;
	}
	wd->sc_dk.dk_openmask = wd->sc_dk.dk_copenmask | wd->sc_dk.dk_bopenmask;

	(void) lockmgr(&wd->sc_dk.dk_openlock, LK_RELEASE, NULL);
	return 0;

 bad2:
	if (wd->sc_dk.dk_openmask == 0) {
		wd->atabus->ata_delref(wd->drvp);
	}
 bad1:
	(void) lockmgr(&wd->sc_dk.dk_openlock, LK_RELEASE, NULL);
	return -error;
}

static int wdclose(
	dev_t dev, 
	int flag, 
	int fmt, 
	struct lwp *l)
{
	struct wd_softc *wd = device_lookup(&wd_cd, WDUNIT(dev));
	int part = WDPART(dev);
	int error;

	if ((error = lockmgr(&wd->sc_dk.dk_openlock, LK_EXCLUSIVE, NULL)) != 0) {
		return -error;
	}

	switch (fmt) {
	case S_IFCHR:
		wd->sc_dk.dk_copenmask &= ~(1 << part);
		break;
	case S_IFBLK:
		wd->sc_dk.dk_bopenmask &= ~(1 << part);
		break;
	}
	wd->sc_dk.dk_openmask = wd->sc_dk.dk_copenmask | wd->sc_dk.dk_bopenmask;

	if (wd->sc_dk.dk_openmask == 0) {
		wd_flushcache(wd, AT_WAIT);

		if (! (wd->sc_flags & WDF_KLABEL)) {
			wd->sc_flags &= ~WDF_LOADED;
		}

		wd->atabus->ata_delref(wd->drvp);
	}

	(void) lockmgr(&wd->sc_dk.dk_openlock, LK_RELEASE, NULL);
	return 0;
}

/*
 *ԸƤհuioϥåƤƤꡢPIO⡼ɤޤDMAԤƼưPIO⡼ɤˤʤȡ
 *        ߤǥѤäƤPIOɤ߹ߤ򤹤Τǡuio̤ˤʤɤ߹߼Ԥˤʤ롣
 */
static int wdread(
	dev_t dev, 
	struct uio *uio,		// ʪ꡼ȥǥֹ椬åȤƤ
	int flags)
{
	return -physio(wdstrategy, NULL, dev, B_READ, (void*) flags, uio);
}

static int wdwrite(
	dev_t dev, 
	struct uio *uio, 
	int flags)
{
	return -physio(wdstrategy, NULL, dev, B_WRITE, (void*) flags, uio);
}

static int wdioctl(
	dev_t dev, 
	u_long xfer, 
	caddr_t addr, 
	int flag, 
	struct lwp *l)
{
	struct wd_softc *wd = device_lookup(&wd_cd, WDUNIT(dev));
	int error = 0, s;

	if ((wd->sc_flags & WDF_LOADED) == 0) {
		return -EIO;
	}

	error = disk_ioctl(&wd->sc_dk, xfer, addr, flag, l);
	if (error != EPASSTHROUGH) {
		return (-error);
	}

	switch (xfer) {
	case DIOCGDINFO:
		*(struct disklabel *)addr = *(wd->sc_dk.dk_label);
		return 0;

	case DIOCGPART:
		((struct partinfo *)addr)->disklab = wd->sc_dk.dk_label;
		((struct partinfo *)addr)->part = &wd->sc_dk.dk_label->d_partitions[WDPART(dev)];
		return 0;
	case DIOCWDINFO:
	case DIOCSDINFO:
	{
		struct disklabel *lp;

		if ((flag & FWRITE) == 0) {
			return -EBADF;
		}

		lp = (struct disklabel *)addr;

		if ((error = lockmgr(&wd->sc_dk.dk_openlock, LK_EXCLUSIVE, NULL)) != 0) {
			goto bad;
		}
		wd->sc_flags |= WDF_LABELLING;

//koko		error = setdisklabel(wd->sc_dk.dk_label, lp, /*wd->sc_dk.dk_openmask : */0, wd->sc_dk.dk_cpulabel);
		if (error == 0) {
			if (wd->drvp->state > RESET) {
				s = splbio();
				wd->drvp->drive_flags |= DRIVE_RESET;
				splx(s);
			}
			if (xfer == DIOCWDINFO) {
//koko				error = writedisklabel(WDLABELDEV(dev), wdstrategy, wd->sc_dk.dk_label, wd->sc_dk.dk_cpulabel);
			}
		}

		wd->sc_flags &= ~WDF_LABELLING;
		(void) lockmgr(&wd->sc_dk.dk_openlock, LK_RELEASE, NULL);
bad:
		return error;
	}

	case DIOCKLABEL:
		if (*(int *)addr) {
			wd->sc_flags |= WDF_KLABEL;
		}
		else {
			wd->sc_flags &= ~WDF_KLABEL;
		}
		return 0;

	case DIOCWLABEL:
		if ((flag & FWRITE) == 0) {
			return EBADF;
		}
		if (*(int *)addr) {
			wd->sc_flags |= WDF_WLABEL;
		}
		else {
			wd->sc_flags &= ~WDF_WLABEL;
		}
		return 0;

	case DIOCGDEFLABEL:
		wdgetdefaultlabel(wd, (struct disklabel *)addr);
		return 0;

	case DIOCGCACHE:
		return wd_getcache(wd, (int *)addr);

	case DIOCSCACHE:
		return wd_setcache(wd, *(int *)addr);

	case DIOCCACHESYNC:
		return wd_flushcache(wd, AT_WAIT);

//	case ATAIOCCOMMAND:
//		/*
//		 * Make sure this command is (relatively) safe first
//		 */
//		if ((((atareq_t *) addr)->flags & ATACMD_READ) == 0 && (flag & FWRITE) == 0) {
//			return (-EBADF);
//		}
//		{
//			struct wd_ioctl *wi;
//			atareq_t *atareq = (atareq_t *) addr;
//			int error1;
//
//			wi = wi_get();
//			wi->wi_softc = wd;
//			wi->wi_atareq = *atareq;
//
//			if (atareq->datalen && atareq->flags & (ATACMD_READ | ATACMD_WRITE)) {
//				wi->wi_iov.iov_base = atareq->databuf;
//				wi->wi_iov.iov_len = atareq->datalen;
//				wi->wi_uio.uio_iov = &wi->wi_iov;
//				wi->wi_uio.uio_iovcnt = 1;
//				wi->wi_uio.uio_resid = atareq->datalen;
//				wi->wi_uio.uio_offset = 0;
//				wi->wi_uio.uio_rw = (atareq->flags & ATACMD_READ) ? B_READ : B_WRITE;
//				wi->wi_uio.uio_vmspace = l->l_proc->p_vmspace;
//				error1 = physio(wdioctlstrategy, &wi->wi_bp, dev, (atareq->flags & ATACMD_READ) ? B_READ : B_WRITE, minphys, &wi->wi_uio);
//			} 
//			else {
//				/* No need to call physio if we don't have any
//				   user data */
//				wi->wi_bp.b_flags = 0;
//				wi->wi_bp.b_data = 0;
//				wi->wi_bp.b_bcount = 0;
//				wi->wi_bp.b_dev = 0;
//				wi->wi_bp.b_proc = l->l_proc;
//				wdioctlstrategy(&wi->wi_bp);
//				error1 = wi->wi_bp.b_error;
//			}
//			*atareq = wi->wi_atareq;
//			wi_free(wi);
//			return(error1);
//		}
//
//	case DIOCAWEDGE:
//		{
//			struct dkwedge_info *dkw = (void *) addr;
//
//			if ((flag & FWRITE) == 0) {
//				return (-EBADF);
//			}
//
//			/* If the ioctl happens here, the parent is us. */
//			strcpy(dkw->dkw_parent, wd->sc_dev.dv_xname);
//			return (dkwedge_add(dkw));
//		}

//	case DIOCDWEDGE:
//		{
//			struct dkwedge_info *dkw = (void *) addr;
//
//		if ((flag & FWRITE) == 0)
//			return (EBADF);
//
//		/* If the ioctl happens here, the parent is us. */
//		strcpy(dkw->dkw_parent, wd->sc_dev.dv_xname);
//		return (dkwedge_del(dkw));
//		}

//	case DIOCLWEDGES:
//		{
//			struct dkwedge_list *dkwl = (void *) addr;
//
//		return (dkwedge_list(&wd->sc_dk, dkwl, l));
//		}

//	case DIOCGSTRATEGY:
//		{
//			struct disk_strategy *dks = (void *)addr;
//
//			s = splbio();
//			strlcpy(dks->dks_name, bufq_getstrategyname(wd->sc_q), sizeof(dks->dks_name));
//			splx(s);
//			dks->dks_paramlen = 0;
//
//			return 0;
//		}
	
//	case DIOCSSTRATEGY:
//		{
//		struct disk_strategy *dks = (void *)addr;
//		struct bufq_state *new;
//		struct bufq_state *old;
//
//		if ((flag & FWRITE) == 0) {
//			return -EBADF;
//		}
//		if (dks->dks_param != NULL) {
//			return -EINVAL;
//		}
//		dks->dks_name[sizeof(dks->dks_name) - 1] = 0; /* ensure term */
//		error = bufq_alloc(&new, dks->dks_name, BUFQ_EXACT|BUFQ_SORT_RAWBLOCK);
//		if (error) {
//			return error;
//		}
//		s = splbio();
//		old = wd->sc_q;
//		bufq_move(new, old);
//		wd->sc_q = new;
//		splx(s);
//		bufq_free(old);

//		return 0;
//		}

	default:
		return ENOTTY;
	}
}

static struct dkdriver wddkdriver = { 
	wdstrategy, 
//	minphys 
};

void wd_shutdown(
	void *arg)
{
	struct wd_softc *wd = arg;
	wd_flushcache(wd, AT_POLL);
}

static int wdprobe(
	struct device *parent, 
	struct cfdata *match, 
	void *aux)
{
	struct ata_device *adev = aux;

	if (adev == NULL) {
		return 0;
	}
	if (adev->adev_bustype->bustype_type != SCSIPI_BUSTYPE_ATA) {
		return 0;
	}

	if (match->cf_loc[ATA_HLCF_DRIVE] != ATA_HLCF_DRIVE_DEFAULT &&
		match->cf_loc[ATA_HLCF_DRIVE] != adev->adev_drv_data->drive) {
		return 0;
	}
	return 1;
}

static void wdattach(
	struct device *parent, 
	struct device *self, 
	void *aux)
{
	struct wd_softc *wd = (void *)self;
	struct ata_device *adev= aux;
	int i, blank;
	char tbuf[41], pbuf[9], c, *p, *q;
	const struct wd_quirk *wdq;

	callout_init(&wd->sc_restart_ch);
//	bufq_alloc(&wd->sc_q, BUFQ_DISK_DEFAULT_STRAT, BUFQ_SORT_RAWBLOCK);
	wd->atabus = adev->adev_bustype;
	wd->openings = adev->adev_openings;
	wd->drvp = adev->adev_drv_data;

	wd->drvp->drv_done = wddone;
	wd->drvp->drv_softc = &wd->sc_dev;

	aprint_naive("\n");

	/* read our drive info */
	if (wd_get_params(wd, AT_WAIT, &wd->sc_params) != 0) {
		aprint_error("\n%s: IDENTIFY failed\n", wd->sc_dev.dv_xname);
		return;
	}

	for (blank = 0, p = wd->sc_params.atap_model, q = tbuf, i = 0; i < (int) sizeof(wd->sc_params.atap_model); i++) {
		c = *p++;
		if (c == '\0') {
			break;
		}
		if (c != ' ') {
			if (blank) {
				*q++ = ' ';
				blank = 0;
			}
			*q++ = c;
		} 
		else {
			blank = 1;
		}
	}
	*q++ = '\0';

	aprint_normal(": <%s>\n", tbuf);

	wdq = wd_lookup_quirks(tbuf);
	if (wdq != NULL) {
		wd->sc_quirks = wdq->wdq_quirks;
	}

	if (wd->sc_quirks != 0) {
		char sbuf[sizeof(WD_QUIRK_FMT) + 64];
		bitmask_snprintf(wd->sc_quirks, WD_QUIRK_FMT, sbuf, sizeof(sbuf));
		aprint_normal("%s: quirks %s\n", wd->sc_dev.dv_xname, sbuf);
	}

	if ((wd->sc_params.atap_multi & 0xff) > 1) {
		wd->sc_multi = wd->sc_params.atap_multi & 0xff;
	} 
	else {
		wd->sc_multi = 1;
	}

	aprint_normal("%s: drive supports %d-sector PIO transfers,", wd->sc_dev.dv_xname, wd->sc_multi);

	/* 48-bit LBA addressing */
	if ((wd->sc_params.atap_cmd2_en & ATA_CMD2_LBA48) != 0) {
		wd->sc_flags |= WDF_LBA48;
	}

	/* Prior to ATA-4, LBA was optional. */
	if ((wd->sc_params.atap_capabilities1 & WDC_CAP_LBA) != 0) {
		wd->sc_flags |= WDF_LBA;
	}

	if ((wd->sc_flags & WDF_LBA48) != 0) {
		aprint_normal(" LBA48 addressing\n");
		wd->sc_capacity =
			((u_int64_t) wd->sc_params.__reserved6[11] << 48) |
			((u_int64_t) wd->sc_params.__reserved6[10] << 32) |
			((u_int64_t) wd->sc_params.__reserved6[9]  << 16) |
			((u_int64_t) wd->sc_params.__reserved6[8]  << 0);
	} 
	else if ((wd->sc_flags & WDF_LBA) != 0) {
		aprint_normal(" LBA addressing\n");
		wd->sc_capacity = ((u_int64_t)wd->sc_params.atap_capacity[1] << 16) | wd->sc_params.atap_capacity[0];
	} 
	else {
		aprint_normal(" chs addressing\n");
		wd->sc_capacity =
			wd->sc_params.atap_cylinders *
			wd->sc_params.atap_heads *
			wd->sc_params.atap_sectors;
	}
	format_bytes(pbuf, sizeof(pbuf), wd->sc_capacity * DEV_BSIZE);
	aprint_normal("%s: %s, %d cyl, %d head, %d sec, %d bytes/sect x %llu sectors\n",
		self->dv_xname, 
		pbuf,
		(wd->sc_flags & WDF_LBA) ? (int)(wd->sc_capacity / (wd->sc_params.atap_heads * wd->sc_params.atap_sectors)) :
		wd->sc_params.atap_cylinders,
		wd->sc_params.atap_heads, 
		wd->sc_params.atap_sectors,
		DEV_BSIZE, 
		(unsigned long long)wd->sc_capacity);

	/*
	 * Initialize and attach the disk structure.
	 */
	wd->sc_dk.dk_driver = &wddkdriver;
	wd->sc_dk.dk_name = wd->sc_dev.dv_xname;
	/* we fill in dk_info later */
	disk_attach(&wd->sc_dk);
	wd->sc_wdc_bio.lp = wd->sc_dk.dk_label;
//	wd->sc_sdhook = shutdownhook_establish(wd_shutdown, wd);
	if (wd->sc_sdhook == NULL) {
		aprint_error("%s: WARNING: unable to establish shutdown hook\n", wd->sc_dev.dv_xname);
	}
#if NRND > 0
	rnd_attach_source(&wd->rnd_source, wd->sc_dev.dv_xname, RND_TYPE_DISK, 0);
#endif

	/* Discover wedges on this disk. */
	dkwedge_discover(&wd->sc_dk);
}

static int wddetach(
	struct device *self, 
	int flags)
{
	struct wd_softc *sc = (struct wd_softc *)self;
	int s;
//	int bmaj;
	int cmaj;
//	int i;
//	int mn;

	/* locate the major number */
//	bmaj = bdevsw_lookup_major(&wd_bdevsw);
	cmaj = cdevsw_lookup_major(&wd_cdevsw);

	/* Nuke the vnodes for any open instances. */
//	for (i = 0; i < MAXPARTITIONS; i++) {
//		mn = WDMINOR(device_unit(self), i);
//		vdevgone(bmaj, mn, mn, VBLK);
//		vdevgone(cmaj, mn, mn, VCHR);
//	}

	/* Delete all of our wedges. */
	dkwedge_delall(&sc->sc_dk);

	s = splbio();

	/* Kill off any queued buffers. */
//	bufq_drain(sc->sc_q);

//	bufq_free(sc->sc_q);
	sc->atabus->ata_killpending(sc->drvp);

	splx(s);

	/* Detach disk. */
	disk_detach(&sc->sc_dk);

	/* Get rid of the shutdown hook. */
//	if (sc->sc_sdhook != NULL) {
//		shutdownhook_disestablish(sc->sc_sdhook);
//	}

#if NRND > 0
	/* Unhook the entropy source. */
	rnd_detach_source(&sc->rnd_source);
#endif

	sc->drvp->drive_flags = 0; /* no drive any more here */

	return (0);
}

static int wdactivate(
	struct device *self, 
	enum devact act)
{
	int rv = 0;

	switch (act) {
	case DVACT_ACTIVATE:
		rv = EOPNOTSUPP;
		break;

	case DVACT_DEACTIVATE:
		/*
		 * Nothing to do; we key off the device's DVF_ACTIVATE.
		 */
		break;
	}
	return (rv);
}

//===================================== PUBLIC =====================================================

CFATTACH_DECL(wd, sizeof(struct wd_softc), wdprobe, wdattach, wddetach, wdactivate);
/*
struct cfattach wd_ca = { 
	"wd", 
	{ ((void *)0), ((void *)0) }, 
	sizeof(struct wd_softc), 
	wdprobe, 
	wdattach, 
	wddetach, 
	wdactivate, 
	0, 
	0 
};
*/

const struct cdevsw wd_cdevsw = {
	wdopen, 
	wdclose, 
	wdread, 
	wdwrite, 
	wdioctl,
	nostop, 
	notty, 
	nopoll, 
	nommap, 
	nokqfilter, 
	D_DISK
};

const struct bdevsw wd_bdevsw = {
	wdopen, 
	wdclose, 
	wdstrategy, 
	wdioctl, 
//	wddump, 
	nodump,
//	wdsize, 
	nosize,
	D_DISK
};
